/*
   hst.c -- PIKT history processing routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/hst.c,v 1.12 1999/12/21 16:00:22 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"
#include "piktd.h"

/* for the given val, do a lookup in the history vals; return a pointer to
   the found record, if any
 */
val *
lookupval (char *key, val *values)
{
	val *vlp;

	foreach (vlp, values)
		if (!strcmp(vlp->key, key))
			return vlp;
	if (HSTTYP & BEGINVAL)
		return valglb;
	else
		return (val *) NULL;
}

/* do a lookup of an alarm, returning a pointer to its data structure
 */
alrm *
lookupalarm (char *name)
{
	alrm *alp;

	foreach (alp, alarms)
		if (!strcmp(alp->name, name))
			return alp;
	return (alrm *) NULL;
}

/* given an input log statement, verify (a) that the log file exists and
   (b) the name of this file matches between what's in the history file and
   what's in the current alarm script
 */
void
verifylog (char *logname)
{
	struct stat fbuf;

	if (stat(logname, &fbuf) != 0) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, TRUE, "in verifylog(), %s, stat() error for %s", ALMNAM, logname);
		HSTERR = TRUE;
	} else if (strcmp(INPUTLOGNAME,logname)) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifylog(), %s, log name in .hst file (%s) doesn't match log name in .alt file (%s)", ALMNAM, logname, INPUTLOGNAME);
		HSTERR = TRUE;
	}
	return;
}

/* verify that the current key var shadows a legitimate var and is in
   the proper sequence
 */
void
verifykey (int vrndx, int arndx)
{

	/* if this vrndx not found in the new history list, or if this
	   vrndx is not next in the key sequence, signify error, else
	   advance to the next key
	 */
	if (!NEWHST(vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifykey(), %s, key index %d in .hst file doesn't match any current history var index", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else if (keycurr && (keycurr->vrndx != vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifykey(), %s, key index %d out of sequence", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else {
		MAXARNDX(vrndx) = maxval(MAXARNDX(vrndx), arndx);
		keycurr = keycurr->next;
	}
	return;
}

/* verify that the current var has not changed, from the history file to
   the present alarm, in various ways that would invalidate use of its
   history vals
 */
void
verifyvar (int vrndx, char *name)
{
	int vx;

	/* if this var name has changed slots, or is a reserved name,
	   or is not a current var, or if this var name has
	   dropped off the new history list, or if this vrndx has
	   dropped off the new history list, or if this varndx was
	   found already in the old history list, signify error
	 */
	if (((vx = varndx(name)) == NIL) || ((vx != vrndx) && (VALTYP(vx) & FIXEDVAL))) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifyvar(), %s, var %s changed index, else reserved var or fn name, else not found", ALMNAM, name);
		HSTERR = TRUE;
	} else if (!NEWHST(vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifyvar(), %s, var index %d in .hst file missing from new history list", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else if (OLDHST(vrndx)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in verifyvar(), %s, duplicate var index %d in history", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else
		OLDHST(vrndx) = TRUE;
	return;
}

/*
void
verifyfld (int vrndx, int arndx, int fld)
{

	/ * if the var's fldnum has changed, signify error * /
	if (!arndx && alarmcurr->vars[vrndx].s && (FLDNUM(vrndx,0) != fld)) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifyfld(), %s, current var fldnum (%d) and history var fldnum (%d) mismatch for var %s", ALMNAM, FLDNUM(vrndx,0), fld, VARNAM(vrndx));
		HSTERR = TRUE;
	} else if (arndx && alarmcurr->vars[vrndx].a && (FLDNUM(vrndx,arndx) != fld)) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifyfld(), %s, current var fldnum (%d) and history var fldnum (%d) mismatch for var %s", ALMNAM, FLDNUM(vrndx,arndx), fld, VARNAM(vrndx));
		HSTERR = TRUE;
	} else
		MAXARNDX(vrndx) = maxval(MAXARNDX(vrndx), arndx);
	return;
}
*/

/* verify that the current var's dattype has not changed; it it has, this
   invalidates use of its history vals
 */
void
verifytyp (int vrndx, int typ)
{

	/* if the var's dattype has changed, signify error */
	if (DATTYP(vrndx) != typ) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifytyp(), %s, current var dattype (%s) and history var dattype (%s) mismatch for var %s", ALMNAM, DATTYP(vrndx) ? "NUMBER" : "STRING", typ ? "NUMBER" : "STRING", VARNAM(vrndx));
		HSTERR = TRUE;
	}
	return;
}

/* verify that the current history val corresponds to either an old or a
   new history var
 */
void
verifyval (int vrndx, int arndx)
{

	/* if this varndx has dropped off the new history list, or if this
	   varndx is missing from the old history list, signify error */
	if (!NEWHST(vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifyval(), %s, val index %d in .hst file missing from new history list", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else if (!OLDHST(vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in verifyval(), %s, val index %d in .hst file missing from old history list", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else
		MAXARNDX(vrndx) = maxval(MAXARNDX(vrndx), arndx);
	return;
}

/* verify that this is not a duplicate global history val
 */
void
verifyglb (int vrndx, int arndx)
{

	verifyval(vrndx, arndx);
	if (! arndx && VALGLB->vals[vrndx].s && VALGLB->VALSCL(vrndx,0)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in verifyglb(), %s, duplicate global history var index %d in history", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else if (VALGLB->vals[vrndx].a && VALGLB->SUBARY(vrndx,arndx) && VALGLB->VALARY(vrndx,arndx)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in verifyglb(), %s, duplicate global history var/ary indices %d/%d in history", ALMNAM, vrndx, arndx);
		HSTERR = TRUE;
	} else
		MAXARNDX(vrndx) = maxval(MAXARNDX(vrndx), arndx);
	return;
}

/* register the current history val, placing it in the current history
   record, and creating a new history record if need be
 */
void
regval (int vrndx, int arndx, char *value)
{
	static char _regval_[] = "regval";
	val *vlp;
	register int i, j;

	/* if there are no registered history values, or if the history
	   value slot for this vrndx is already filled, malloc and link a
	   new history record
	 */
	if ((valcurr == VALGLB) || (! arndx && valcurr->vals[vrndx].s && valcurr->VALSCL(vrndx,0)) || (valcurr->vals[vrndx].a && valcurr->SUBARY(vrndx,arndx) && valcurr->VALARY(vrndx,arndx))) {
		vlp = (val *) mymalloc(sizeof(val), _regval_);
		vlp->prev = valcurr;
		vlp->next = (val *) NULL;
		vlp->key = (char *) NULL;
		for (i=0; i<=MAXVARS; i++) {
			switch (VARTYP(i)) {
				case SCALAR:
					vlp->vals[i].s = (chrscl *) mymalloc(sizeof(chrscl), _regval_);
					vlp->vals[i].s->val[0] = (char *) NULL;
					break;
				case ARRAY:
					vlp->vals[i].a = (chrary *) mymalloc(sizeof(chrary), _regval_);
					for (j=0; j<MAXSUBARY; j++) {
						vlp->vals[i].a->subary[j] = (chrsubary *) NULL;
					}
					break;
				case NIL:
				default:
					vlp->vals[i].s = (chrscl *) NULL;
					vlp->vals[i].a = (chrary *) NULL;
					break;
			}
		}
		valcurr->next = vlp;
		valcurr = vlp;
	}
	/* fill the slot */
	SETVAL(valcurr,vrndx,arndx,value);
	MAXARNDX(vrndx) = maxval(MAXARNDX(vrndx), arndx);
	return;
}

/* for all alarms, and all lookup keys, generate key vals for all
   history records
 */
void
makekeys (void)
{
	alrm *alp;
	val *vlp;
	int concat;
	char keystr[MAXLINE+1];
	ky *kp;

	keystr[0] = '\0';
	foreach (alp, alarms) {
		if (!alp->hsterr) {
			foreach (vlp, alp->values->next->next->next) {
				concat = FALSE;
				foreach (kp, alp->keys) {
					switch (alp->vartype[kp->vrndx]) {
						case SCALAR:
							if (vlp->vals[kp->vrndx].s && vlp->VALSCL(kp->vrndx,0)) {
								if (!concat) {
									strcpy(keystr,vlp->VALSCL(kp->vrndx,0));
									concat = TRUE;
								} else {
									strcat(keystr,":");
									strcat(keystr,vlp->VALSCL(kp->vrndx,0));
								}
							}
							break;
						case ARRAY:
							if (vlp->vals[kp->vrndx].a && vlp->SUBARY(kp->vrndx,kp->arndx) && vlp->VALARY(kp->vrndx,kp->arndx)) {
								if (!concat) {
									strcpy(keystr,vlp->VALARY(kp->vrndx,kp->arndx));
									concat = TRUE;
								} else {
									strcat(keystr,":");
									strcat(keystr,vlp->VALARY(kp->vrndx,kp->arndx));
								}
							}
							break;
						case NIL:
							break;
						default:
							logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in makekeys(), %s, unknown vartype for var %d", alp->name, kp->vrndx);
							alp->hsterr = TRUE;
							break;
					}
				}
				vlp->key = mystrdup(keystr, "makekeys");
			}
		}
	}
	return;
}

/* free vals in the val struct, either when disposing of a history val
   that's no longer needed, or when assigning new valnows
 */
void
freevals (val **vlp, int type, int terminal)
{
	register int i, j, k;
	int maxsubary;
	int maxsubndx;

	for (i=1; i<=MAXVARS; i++) {
		if (VALTYP(i) & type) {
			switch (VARTYP(i)) {
				case SCALAR:
					if ((*vlp)->vals[i].s) {
						if ((*vlp)->VALSCL(i,0)) {
							if (isdebug)
								logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "freeing val %s", (*vlp)->VALSCL(i,0));
							free((*vlp)->VALSCL(i,0));
							(*vlp)->VALSCL(i,0) = (char *)NULL;
						}
						if (terminal)
							free((*vlp)->vals[i].s);
					}
					break;
				case ARRAY:
					if ((*vlp)->vals[i].a) {
						maxsubary = SA(MAXARNDX(i));
						for (j=0; j<=maxsubary; j++) {
							if ((*vlp)->vals[i].a->subary[j]) {
								if (j == maxsubary)
									maxsubndx = SN(MAXARNDX(i));
								else
									maxsubndx = 255;
								for (k=0; k<=maxsubndx; k++) {
									if ((*vlp)->vals[i].a->subary[j]->val[k]) {
										if (isdebug)
											logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "freeing val %s", (*vlp)->vals[i].a->subary[j]->val[k]);
										free((*vlp)->vals[i].a->subary[j]->val[k]);
										(*vlp)->vals[i].a->subary[j]->val[k] = (char *)NULL;
									}
								}
								free((*vlp)->vals[i].a->subary[j]);
								(*vlp)->vals[i].a->subary[j] = (chrsubary *) NULL;
							}
						}
						if (terminal)
							free((*vlp)->vals[i].a);
					}
					break;
				case NIL:
					break;
				default:
					break;
			}
		}
	}
	return;
}

/* free a history value from the history list
 */
void
freevalhst (val *vlp)
{
	static val *tvlp;
	val *pvlp;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "entering freevalhst()");
		dumpvals(VALUES);
	}
	tvlp = vlp;
	if ((pvlp = vlp->prev)) {
		pvlp->next = vlp->next;
		if (vlp->next)
			vlp->next->prev = pvlp;
	} else
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in freevalhst(), %s, pvlp = vlp->prev failure", ALMNAM);
	FREE(tvlp->key);
	freevals(&tvlp, FIXEDVAL|FREEVAL, FALSE);
	free(tvlp);
	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "leaving freevalhst()");
		dumpvals(VALUES);
	}
	return;
}

/* free any remaining history values from the history list not previously
   freed in freevalhst
 */
void
freevalhsts (val **values)
{
	val *vlp;
	static val *tvlp;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "entering freevalhsts()");
		dumpvals(VALUES);
	}
	vlp = *values;
	while (vlp) {
		tvlp = vlp;
		vlp = vlp->next;
		FREE(tvlp->key);
		freevals(&tvlp, FIXEDVAL, FALSE);
		free(tvlp);
	}
	*values = (val *) NULL;
	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "leaving freevalhsts()");
		dumpvals(VALUES);
	}
	return;
}
